/* $Id: vrfy.C,v 1.12 2004/11/18 23:03:23 dm Exp $ */

/*
 *
 * Copyright (C) 2004 David Mazieres (dm@uun.org)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "asmtpd.h"

class mailserv {
public:
  typedef callback<void, str, int>::ref chatcb_t;
private:
  ptr<aios> aio;
  
  bool lineconsistent (strbuf sb, str line);
  void concb (str line, int err);
  void helocb (bool ehlo, str line, int err);
  void chatcb (ref<vec<str> > cv, chatcb_t cb, strbuf sb, str line, int err);
  void unsolicited (str line, int err);
  void touch ();

protected:
  bool busy;
  bool noop;
  void idle ();
  virtual void fail (bool temp, str error);
  virtual void ready () { busy = false; }
public:
  static u_int nmailserv;
  static str heloname () { return opt->hostname; }

  const str name;
  ihash_entry<mailserv> hlink;
  tailq_entry<mailserv> llink;

  mailserv (str n, int fd);
  virtual ~mailserv ();
  void helo ();
  void chat (const vec<str> &cmds, chatcb_t cb);
  static mailserv *lookup (str name);
};

struct mailserv_readycb : mailserv {
  timecb_t *tmo;
  cbb::ptr cb;

  mailserv_readycb (str n, int fd)
    : mailserv (n, fd), tmo (NULL) {}
  ~mailserv_readycb () {
    timecb_remove (tmo);
    if (cb) (*cb) (false);
  }
  void ready () {
    idle ();
    busy = true;
    tmo = delaycb (opt->vrfy_delay, wrap (this, &mailserv_readycb::ready_2));
  }
  void ready_2 () {
    tmo = NULL;
    busy = false;
    cbb c (cb);
    cb = NULL;
    (*c) (true);
  }
};

static ihash<const str, mailserv, &mailserv::name, &mailserv::hlink> mstab;
static tailq<mailserv, &mailserv::llink> msq;
u_int mailserv::nmailserv;

mailserv::mailserv (str n, int fd)
  :  aio (aios::alloc (fd)), busy (true), name (mytolower (n))
{
  nmailserv++;
  mstab.insert (this);
  msq.insert_tail (this);
  if (opt->debug_smtpc)
    aio->setdebug (strbuf ("%s (R%d)", name.cstr (), fd));
  aio->settimeout (opt->client_timeout);
  // XXX - this will, of course, fail to close mailservs behind slow clients
  while (nmailserv > opt->max_revclients && msq.first && !msq.first->busy)
    delete msq.first;
}
mailserv::~mailserv ()
{
  nmailserv--;
  mstab.remove (this);
  msq.remove (this);
}
void
mailserv::helo ()
{
  aio->readline (wrap (this, &mailserv::concb));
}
void
mailserv::fail (bool temp, str error)
{
  if (opt->debug_smtpc)
    warn << name << ": " << error << "\n";
  delete this;
}

void
mailserv::concb (str line, int err)
{
  if (!line || line.len () < 4)
    fail (true, err ? strerror (err) : "eof");
  else if (line[3] == '-')
    aio->readline (wrap (this, &mailserv::concb));
  else if (line[0] != '2')
    fail (line[0] != '5', substr (line, 4));
  else {
    aio << "EHLO " << heloname () << "\r\n";
    aio->readline (wrap (this, &mailserv::helocb, true));
  }
}

void
mailserv::helocb (bool ehlo, str line, int err)
{
  if (!line || line.len () < 4)
    fail (true, err ? strerror (err) : "eof");
  else if (line[3] == '-')
    aio->readline (wrap (this, &mailserv::helocb, ehlo));
  else if (line[0] == '5' && ehlo) {
    aio << "HELO " << heloname () << "\r\n";
    aio->readline (wrap (this, &mailserv::helocb, false));
  }
  else if (line[0] != '2')
    fail (line[0] != '5', substr (line, 4));
  else {
    busy = false;
    ready ();
    idle ();
  }
}

void
mailserv::chat (const vec<str> &cmds, mailserv::chatcb_t cb)
{
  assert (!busy);
  busy = true;
  aio->readcancel ();
  touch ();

  ref<vec<str> > cv = New refcounted<vec<str> > (cmds);
  if (cv->empty ()) {
    busy = false;
    (*cb) (NULL, 0);
    idle ();
  }
  else {
    noop = !strncasecmp (cv->front (), "noop", 4);
    aio << cv->pop_front () << "\r\n";
    aio->readline (wrap (this, &mailserv::chatcb, cv, cb, strbuf ()));
  }
}

bool
mailserv::lineconsistent (strbuf sb, str line)
{
  static rxx resprx ("^(\\d\\d\\d)(-| )(.*)$");
  if (!line || !resprx.match (line))
    return false;
  if (sb.tosuio ()->resid () > 3) {
    char buf[3];
    sb.tosuio ()->copyout (buf, 3);
    if (memcmp (buf, line.cstr (), 3))
      return false;
  }
  else if (!sb.tosuio ()->resid ())
    sb << resprx[1] << "-server " << name << " reports:\r\n";
  sb << line << "\r\n";
  return true;
}
void
mailserv::chatcb (ref<vec<str> > cv, chatcb_t cb, strbuf sb,
		  str line, int err)
{
  if (!lineconsistent (sb, line)) {
#if 1
    if (line) {
      warn ("inconsistent line -------------------------------------------\n");
      warn << sb << line << "\n";
      warn ("-------------------------------------------------------------\n");
    }
#endif
    fail (true, err ? strerror (err) : "eof");
    (*cb) (NULL, err);
  }
  else if (line[3] == '-')
    aio->readline (wrap (this, &mailserv::chatcb, cv, cb, sb));
  else if ((!noop && line[0] != '2') || cv->empty ()) {
    busy = false;
    (*cb) (sb, 0);
    idle ();
  }
  else {
    sb.tosuio ()->clear ();
    noop = !strncasecmp (cv->front (), "noop", 4);
    aio << cv->pop_front () << "\r\n";
    aio->readline (wrap (this, &mailserv::chatcb, cv, cb, sb));
  }
}

void
mailserv::idle ()
{
  if (!busy) {
    aio->readcancel ();
    aio->readline (wrap (this, &mailserv::unsolicited));
  }
}

void
mailserv::unsolicited (str line, int err)
{
  fail (true, "eof");
}

void
mailserv::touch ()
{
  msq.remove (this);
  msq.insert_tail (this);
}

mailserv *
mailserv::lookup (str name)
{
  name = mytolower (name);
  for (mailserv *msp = mstab[name]; msp; msp = mstab.nextkeq (msp))
    if (!msp->busy)
      return msp;
  return NULL;
}

class mxconnect {
  typedef callback<void, mailserv *, int, ptr<mxlist> >::ref cb_t;

  str name;
  cb_t cb;
  int n;
  ptr<mxlist> mxl;
  int fallthrough_error;

  mxconnect (str nn, cb_t c)
    : name (nn), cb (c), n (-1), fallthrough_error (ARERR_NXREC) {}
  void start () {
    dns_mxbyname (name, wrap (this, &mxconnect::mxcb));
  }
  void mxcb (ptr<mxlist> m, int dnserr) {
    mxl = m;
    //printmxlist (name, mxl, dnserr);
    if (mxl)
      trymx ();
    else if (dnserr == ARERR_NXREC)
      trya (name);
    else {
      (*cb) (NULL, dnserr, mxl);
      delete this;
    }
  }
  void trymx () {
    if (!mxl || implicit_cast<unsigned> (++n) >= mxl->m_nmx) {
      (*cb) (NULL, fallthrough_error, mxl);
      delete this;
    }
    else if (mailserv *msp = mailserv::lookup (mxl->m_mxes[n].name)) {
      (*cb) (msp, 0, mxl);
      delete this;
    }
    else
      trya (mxl->m_mxes[n].name);
  }
  void trya (str a) {
    /* XXX - probably want a smaller timeout */
    dns_hostbyname (a, wrap (this, &mxconnect::tryip, a));
  }
  void tryip (str a, ptr<hostent> h, int err) {
    if (!h) {
      if (fallthrough_error) {
	if (dns_tmperr (err))
	  fallthrough_error = err;
	else if (!dns_tmperr (fallthrough_error))
	  fallthrough_error = err;
      }
      errno = dns_tmperr (err) ? EAGAIN : ENOENT;
      tcpcb (a, -1);
      return;
    }
    fallthrough_error = 0;
    u_int32_t ip = ntohl (((in_addr *) h->h_addr)->s_addr);
    if (ip >> 24 == 0 || ip >> 24 == 127 || ip == 0xffffffff) {
      errno = EACCES;
      tcpcb (a, -1);
      return;
    }
    tcpconnect (*(in_addr *) h->h_addr, 25, wrap (this, &mxconnect::tcpcb, a));
  }
  void tcpcb (str a, int fd) {
    if (fd < 0) {
      if (opt->debug_smtpc)
	warn << "mxconnect: " << a << " for " << name << ": "
	     << strerror (errno) << "\n";;
      trymx ();
    }
    else {
      if (opt->debug_smtpc)
	warn << "mxconnect: " << a << " for " << name << ": success (fd "
	     << fd << ")\n";
      mailserv_readycb *msp = New mailserv_readycb (a, fd);
      msp->cb = wrap (this, &mxconnect::ready, msp);
      msp->helo ();
    }
  }
  void ready (mailserv *msp, bool ok) {
    if (ok) {
      (*cb) (msp, 0, mxl);
      delete this;
    }
    else
      trymx ();
  }

public:
  static void alloc (str n, cb_t c)
    { (New mxconnect (n, c))->start (); }
};

static void
vrfy_3 (str addr, vrfycb_t cb, ptr<mxlist> mxl, str line, int err)
{
  if (line && line[0] == '2')
    (*cb) (NULL, mxl);
  else if (line && (line[0] == '4' || line[0] == '5')) {
    if (line.len () > 4)
      (*cb) (line, mxl);
    else if (line[0] == '5')
      (*cb) ("550 bad user\r\n", mxl);
    else
      (*cb) ("451 bad user\r\n", mxl);
  }
  else {
    str relay = extract_domain (addr);
    if (err)
      (*cb) (strbuf ("451 %s: %s\r\n", relay.cstr (), strerror (err)), mxl);
    else
      (*cb) (strbuf ("451 %s: SMTP protocol failure\r\n", relay.cstr ()), mxl);
  }
}
static void
vrfy_2 (str addr, in_addr cli, vrfycb_t cb,
	mailserv *msp, int dnserr, ptr<mxlist> mxl)
{
  if (dnserr)
    (*cb) (strbuf ("%03d %s\r\n", dns_tmperr (dnserr) ? 451 : 553,
		   dns_strerror (dnserr)), NULL);
  else if (!msp)
    (*cb) (strbuf () << "451 could not connect to server for "
	   << extract_domain (addr) << "\r\n", NULL);
  else {
    vec<str> cmds;
    cmds.push_back ("RSET");
    cmds.push_back (strbuf ("NOOP %s is sending mail from <%s>; if forged, "
			    "consider an SPF record (http://spf.pobox.com/)",
			    inet_ntoa (cli), addr.cstr ()));
    cmds.push_back ("MAIL FROM:<>");
    cmds.push_back (strbuf () << "RCPT TO:<" << addr << ">");
    msp->chat (cmds, wrap (vrfy_3, addr, cb, mxl));
  }
}
void
vrfy (str addr, in_addr cli, vrfycb_t cb)
{
  str relay = extract_domain (addr);
  if (!relay)
    (*cb) ("501 syntax error\r\n", NULL);
  else
    mxconnect::alloc (relay, wrap (vrfy_2, addr, cli, cb));
}
